/*------------------------------------------------------------------------------
  Purpose:     DBQuery   
  Parameters:  pcTables         - Buffer names. 
               pcPhysicalTables - Physical names.          
  Notes:      The concept of properties is not directly supported in the current
              ABL. It can be implemented with getter and setters wrapping 
              private or protected data members. 
              This sample have a somewhat inconsistent use of getters and 
              setters and public members for what should be implemented 
              according to the implementers chosen strategy and convention for 
              properties.           
------------------------------------------------------------------------------*/
CLASS com.openedge.framework.util.query.DBQuery /* inherits Query ... */ :
  /* Pretend properties (must be controlled) */ 
  DEFINE PUBLIC VARIABLE Tables         AS CHAR NO-UNDO.
  DEFINE PUBLIC VARIABLE BaseQuery      AS CHARACTER  NO-UNDO.

  DEFINE PROTECTED VARIABLE PhysicalTables AS CHARACTER  NO-UNDO.
  DEFINE PROTECTED VARIABLE QueryHandle     AS HANDLE NO-UNDO.
  
  /* ADM service adapter uses commas as delimiter for each temp-table
     not each buffer, so we may need a property if this does not fit other
     clients (one can override the adapter to use another delimiter 
     or of course just do the conversion in the adapter which is
     what it's for) */      
  DEFINE PROTECTED  VARIABLE mcRowidDlm        AS CHARACTER  NO-UNDO INIT ";".

  CONSTRUCTOR PUBLIC DBQuery(pcPhysicalTables AS CHAR,pcTables AS CHAR): 
    DEFINE VARIABLE iBuffer AS INTEGER    NO-UNDO.
    DEFINE VARIABLE hBuffer AS HANDLE     NO-UNDO.
    /* We operate on Tables, but have it as optional parameter ! */
    IF pcTables = '' THEN
      pcTables = pcPhysicalTables.
    ASSIGN
      PhysicalTables = pcPhysicalTables
      Tables         = pcTables.

    CREATE QUERY QueryHandle.
    DO iBuffer = 1 TO NUM-ENTRIES(pcPhysicalTables):
      CREATE BUFFER hBuffer 
        FOR TABLE ENTRY(iBuffer,pcPhysicalTables) BUFFER-NAME ENTRY(iBuffer,pcTables).
      QueryHandle:ADD-BUFFER(hBuffer).
    END.

  END CONSTRUCTOR.

  DESTRUCTOR PUBLIC DBQuery () :
    /* delete the handles  */
    DEFINE VARIABLE iBuffer AS INTEGER    NO-UNDO.
    DEFINE VARIABLE h       AS HANDLE     NO-UNDO EXTENT 18.
    DEFINE VARIABLE iNum    AS INTEGER    NO-UNDO.

    IF VALID-HANDLE(QueryHandle) THEN
    DO:
      iNum = QueryHandle:NUM-BUFFERS.
      /* we loose num-buffers on first delete */
      DO iBuffer = 1 TO iNum:
        h[iBuffer] = QueryHandle:GET-BUFFER-HANDLE(iBuffer).
      END.
      DO iBuffer = 1 TO iNum: 
        DELETE OBJECT h[iBuffer].
      END.
      DELETE OBJECT QueryHandle NO-ERROR.
    END.

  END DESTRUCTOR.

  METHOD PROTECTED LOGICAL prepare ():
    RETURN QueryHandle:QUERY-PREPARE(currentQuery()).     
  END.
  
  METHOD PROTECTED LOGICAL resetQuery ():
    RETURN QueryHandle:QUERY-PREPARE(defaultQuery()).     
  END.
  
  METHOD PROTECTED CHARACTER currentQuery ():
  /*------------------------------------------------------------------------------
    Purpose:     Return the current query for query manipulation and prepare.
                 returns the default if not yet prepared.
  ------------------------------------------------------------------------------*/
    DEFINE VARIABLE cQueryString AS CHARACTER  NO-UNDO.
    
    IF QueryHandle:PREPARE-STRING > '' THEN
      RETURN QueryHandle:PREPARE-STRING.

    RETURN defaultQuery().

  END METHOD. 

  METHOD PUBLIC CHARACTER defaultQuery ():
  /*------------------------------------------------------------------------------
    Purpose:     Return the BaseQuery or build a default query. 
  ------------------------------------------------------------------------------*/
    DEFINE VARIABLE iBuffer     AS INTEGER    NO-UNDO.
    DEFINE VARIABLE cPrepare    AS CHARACTER  NO-UNDO.
    DEFINE VARIABLE cBuffer     AS CHARACTER  NO-UNDO.
    DEFINE VARIABLE cParent     AS CHARACTER  NO-UNDO.
    DEFINE VARIABLE cMode       AS CHARACTER  NO-UNDO.
    DEFINE VARIABLE cKeyTable   AS CHARACTER  NO-UNDO.

    IF BaseQuery <> '' THEN
      RETURN BaseQuery.
    
    /* assume the first table in the definition is the main table that 
       the others join to and that need to have 'EACH' if joined from
       one of the other tables */
    cKeyTable = ENTRY(1,Tables).
    
    /* If there's more than one buffer than add them, just assuming that
       an OF relationship to the first table in tables will properly relate them. */
    cPrepare = " FOR EACH " + QueryHandle:GET-BUFFER-HANDLE(1):NAME + " NO-LOCK ".
    DO iBuffer = 2 TO QueryHandle:NUM-BUFFERS:

      ASSIGN 
        cBuffer  = QueryHandle:GET-BUFFER-HANDLE(iBuffer):NAME
        cParent  = IF cKeyTable = cBuffer THEN 
                     QueryHandle:GET-BUFFER-HANDLE(1):NAME
                   ELSE  
                     QueryHandle:GET-BUFFER-HANDLE(cKeyTable):NAME
        cMode    = IF cKeyTable = cBuffer THEN 'EACH' ELSE 'FIRST'
        cPrepare = cPrepare 
                 + ", " + cMode + " " + cBuffer + " OF " +  cParent + " NO-LOCK".

    END.   /* DO iBuffer = 2 */
    cPrepare = cPrepare + ' INDEXED-REPOSITION'.  
    RETURN cPrepare.
  END.

  METHOD PUBLIC CHARACTER columnValue (pcColumn AS CHAR):
    DEFINE VARIABLE cBuffer AS CHARACTER  NO-UNDO.
    DEFINE VARIABLE cColumn AS CHARACTER  NO-UNDO.
    DEFINE VARIABLE cValue  AS CHARACTER  NO-UNDO.
    DEFINE VARIABLE hBuffer AS HANDLE     NO-UNDO.

    ASSIGN
      cBuffer = ENTRY(1,pcColumn,'.')
      cColumn = ENTRY(2,pcColumn,'.')  
      hBuffer = QueryHandle:GET-BUFFER-HANDLE(cBuffer).
    IF hBuffer:AVAIL THEN
    DO:
      /* extent support could be added by <field>[i] as param, 
        but this used for keys though */
      cValue = hBuffer:BUFFER-FIELD(cColumn):BUFFER-VALUE(0). 
      /* this string is for transport of values, so return unknown as string
       (assuming '?' never is a value in a progress DB...)*/        
      RETURN IF cValue <> ? THEN cValue ELSE '?'.
    END.
    RETURN ?. /* not avail*/ 
  END METHOD.


  /* order neutral position that is safe if tables changes order 
     passed back to setPosition  
     note that one also need all rowids when the unique table is not 
     the first  */
  METHOD PUBLIC CHARACTER getPosition ():
     DEFINE VARIABLE iBuffer AS INTEGER    NO-UNDO.
     DEFINE VARIABLE cBuffer AS CHARACTER  NO-UNDO.
     DEFINE VARIABLE cPosition AS CHARACTER  NO-UNDO.

     DO iBuffer = 1 TO NUM-ENTRIES(Tables):
       ASSIGN
         cBuffer = ENTRY(iBuffer,Tables)    
         cPosition = cPosition 
                   + (IF iBuffer = 1 THEN '' ELSE mcRowidDlm)
                   +  STRING(QueryHandle:GET-BUFFER-HANDLE(cBuffer):ROWID).

     END.

     /* blank for no avail */
     RETURN IF cPosition = ? THEN '' ELSE cPosition.
  END METHOD.
 
  /* set position as returned from getPosition */
  METHOD PUBLIC LOGICAL setPosition (pcPosition AS CHAR):
    DEFINE VARIABLE iBuffer    AS INTEGER                       NO-UNDO.
    DEFINE VARIABLE iTable     AS INTEGER                       NO-UNDO.
    DEFINE VARIABLE rPosition  AS ROWID   EXTENT 18 /*MAXJOIN*/ NO-UNDO.
    DEFINE VARIABLE lOk        AS LOGICAL                       NO-UNDO.

    DO iBuffer = 1 TO QueryHandle:NUM-BUFFERS:
      ASSIGN 
        iTable = LOOKUP(QueryHandle:GET-BUFFER-HANDLE(iBuffer):NAME,Tables)
        rPosition[iTable] = TO-ROWID(ENTRY(iBuffer,pcPosition,mcRowidDlm)).
    END.
    lOk = QueryHandle:REPOSITION-TO-ROWID(rPosition) NO-ERROR.
    IF lOk THEN 
      QueryHandle:GET-NEXT.
    RETURN lOk.
  END METHOD.
 
END.

