/*------------------------------------------------------------------------------
 File        : daOrder order data access   
 Description : Example/prototype code  
 Notes       : Illustrates solution for two issues 
               1. Logical data wants real key  
               2. Database defines view-as with hard coded list for creditcard
               we return it normalized as table with extrenal keys (pretend).         
 ------------------------------------------------------------------------------*/

/* ***************************  Definitions  ************************** */
/* defines dataset include and loads daSupport as super */ 
{support/daEntity.i Order} 

define data-source dOrder for Order.
define data-source dOrderLine for OrderLine.
define data-source dItem for Item.

define variable gcCreditCardKeys as character no-undo
  init 'AmEx,MC,Visa,Disc'. 
define variable gcCreditCardNames as character no-undo
  init 'American Express,Master Card,Visa,Discover Card'. 
 
/* ************ Functions ****************************************/           
/* call backs for query mapping */
function sourceDefaultQuery returns character  
  (pcTable as char):
  case pcTable:
    when 'eOrder' then 
      return 'For each Order indexed-reposition'.
    when 'eOrderLine' then 
      return 'For each OrderLine indexed-reposition'.
  end.    
end.

function sourceColumn returns character  
  (pcTable as char,
   pcColumn as char ):
  define variable cMap  as character no-undo.
  case pcTable:
    when 'eOrder' then
    do: 
      cMap = data-source dorder:get-dataset-buffer:data-source-complete-map.
    end.  
    when 'eOrderLine' then
    do: 
      cMap = data-source dorderline:get-dataset-buffer:data-source-complete-map.
    end.  
  end.
  return entry(lookup(pccolumn,cMap) + 1,cMap).
end.

function creditCardName returns character private 
  (pcKey as character):
  return entry(lookup(pcKey,gcCreditCardKeys),gcCreditCardNames).
end.

function creditCardKey returns character private 
  (pcName as character):
  return entry(lookup(pcName,gcCreditCardNames),gcCreditCardKeys).
end.

function errorString returns character private 
  (pcString as char):
  if pcString > '' then 
    return pcString.
  else    
    return error-status:get-message(1). 
end.
   
/* ************ Procedures ****************************************/        

procedure attachDataSources:
  define input-output parameter dataset for dsOrder.
 /* this should probably be handled elsewhere typically not an issue on a server 
    but this test is being run with local connection */
  temp-table eOrder:tracking-changes = false. 
  temp-table eOrderLine:tracking-changes = false. 
  temp-table eCreditCard:tracking-changes = false. 
  temp-table eItem:tracking-changes = false. 
  
  buffer eOrder:attach-data-source(data-source dOrder:handle,'','CreditCard').        
  buffer eOrderLine:attach-data-source(data-source dOrderLine:handle).
  buffer eItem:attach-data-source(data-source dItem:handle).  
end.   

/* Called from fetchData */ 
procedure defineReadEvents : 
  define input-output parameter dataset for dsOrder.
 
  buffer eOrder:set-callback-procedure
            ("AFTER-ROW-FILL", "eOrderAfterRowFill", target-procedure).
  
  buffer eCreditCard:set-callback-procedure
            ("BEFORE-FILL", "fillCreditCard", target-procedure).     

end.
procedure saveChanges:
/*------------------------------------------------------------------------------
  Purpose:  Save changes to physical storage                
Parameters: dataset dsOrder 
     Notes: This example of data access saves all in one transaction, and
            only adds error to failed record (and dataset)
-----------------------------------------------------------------------------*/
  define input-output parameter dataset for dsOrder.

  run attachDataSources (input-output dataset dsOrder by-reference). 
  /* traverse before records to get deletes */  
  OrderTrans: 
  do transaction:
    /* do orderline deletes first */
    for each eOrderLineBefore where buffer eOrderLineBefore:row-state = row-deleted:             
      buffer eOrderLineBefore:save-row-changes() no-error.
      if buffer eOrderLineBefore:error then 
      do:
        buffer eOrderLineBefore:error-string = errorString(return-value).
        UNDO OrderTrans,LEAVE OrderTrans.
      end.
    end. /* each eorderlinebefore = delete */
      
    for each eOrderBefore:
      if buffer eOrderBefore:row-state <> row-created then
      do:
        find order where order.ordernum = eOrderBefore.ordernum exclusive no-error.
        if not available order then
        do:
          assign 
            buffer eOrderBefore:error = true
            buffer eOrderBefore:error-string = 
               'Order ' + string(eOrderBefore.ordernum) 
               + (if locked order then 'is locked by another user/session'
                  else ' does not exist (anymore)')  
            buffer eOrderBefore:data-source-modified = true.
            UNDO OrderTrans,LEAVE OrderTrans.
        end. /* not avail or locked order */
        /* optimistic lock check for transformed data */
        if Order.Creditcard <> creditCardName(eOrderBefore.CreditCard) then 
        do:
          assign 
            buffer eOrderBefore:error = true
            buffer eOrderBefore:data-source-modified = true.
          run eOrderAfterRowFill. /* refresh */
          UNDO OrderTrans,LEAVE OrderTrans. 
        end.    
      end.
      buffer eOrderBefore:save-row-changes(1,'ordernum') no-error.      
      if buffer eOrderBefore:error then
      do: 
        buffer eOrderBefore:error-string = errorString(return-value).
         /* refresh transformated data if lock conflict */
        if buffer eOrderBefore:data-source-modified then
          run eOrderAfterRowFill.
          
        UNDO OrderTrans,LEAVE OrderTrans.
      end.
      if buffer eOrderBefore:row-state <> row-deleted then
      do:  
        /* transform to internal format */
        find eOrder where rowid(eOrder) = buffer eorderbefore:after-rowid.         
        find order where order.ordernum = eOrder.ordernum exclusive.
        Order.CreditCard = creditCardName(eOrder.CreditCard).
      end.           
    end. /* for each eOrderBefore */
    
    /* save orderline changes  */ 
    for each eOrderLineBefore where buffer eOrderLineBefore:row-state <> row-deleted:             
      buffer eOrderLineBefore:save-row-changes() no-error.
      if buffer eOrderLineBefore:error then 
      do:
        buffer eOrderLineBefore:error-string = errorString(return-value).
        UNDO,NEXT.
      end.       
    end. /* for each eOrderLineBefore <> delete */
  end. /* transaction */ 
  run detachDataSources (input-output dataset dsOrder by-reference). 
end. /* saveChanges */

procedure fillCreditCard:
  define input parameter dataset for dsOrder.
  /* Physical data do not normalize creditcard info, use internal key 
     and depend on view-as. This illustrates a solution for two issues 
     1. Logical data wants real key (pretend: data is from external register) 
     for interoperability with other (services...) 
     2. A UI cannot be expected to support view-as and should not need to 
        hardcode list, but should be able to bind to a table          */    
  define variable iCard  as integer   no-undo.
  do iCard = 1 to num-entries(gcCreditCardKeys):
    create eCreditCard.
    assign
      eCreditCard.CreditCard     = entry(iCard,gcCreditCardKeys)
      eCreditCard.CreditCardName = entry(iCard,gcCreditCardNames).
  end.           
end.  /* fillCreditCard */    

procedure eOrderAfterRowFill:  
  define input parameter dataset for dsOrder.
  /* transform internal physical key to external key  
     This is not meant to indicate that this is desirable, but highlight that 
     the DA layer can overcome physical limitations  
     - keep in mind that any row level hook add cost to the fill */       
  eOrder.CreditCard = creditCardKey(Order.CreditCard).
end.  /* eOrderAfterRowFill */   


       