/*------------------------------------------------------------------------------
 File        : support/dasupport.p    
 Description : data access super 
 Notes       : 
             - fetchData supports conditional retrieval of tables
             - All request methods are statefree (attaches and detaches sources) 
               and passes dataset around  by-reference. (no get or set)
             - query is only supported for the first table   
 ------------------------------------------------------------------------------*/

/* ***************************  Definitions  ************************** */
using com.openedge.framework.util.query.*.

/* ************ Functions ****************************************/        
/* To be used by generic super destroy  */
function getObjectType returns character ( ) :
  if target-procedure = this-procedure then 
    return "super". 
  else 
    return "DataAccessObject".                                            
end.
 

procedure fetchData:
/*------------------------------------------------------------------------------
  Purpose: Fetch data from the physical storage.
               
Parameters:  
     phDataSet      - Prodataset handle (NOT dataset-handle)
     pcContext      - context    
     pcTables       - Tables for which to retrieve data or definitions
     pcQueries      - List of queries that corresponds to the requested tables
     pcRequest      - The actual request for the tables   
     piNumRecords   - Batch sizes
 
   Notes:  uses static data sources   
 -----------------------------------------------------------------------------*/
  define input        parameter pcTables       as character no-undo.
  define input        parameter pcQueries      as character no-undo.
  define input        parameter pcRequest      as character no-undo.
  define input        parameter piBatchSize    as integer   no-undo.
  define input-output parameter pcContext      as character no-undo.
  define output       parameter dataset-handle phDataset.
  
  define variable iBuffer      as integer   no-undo.
  define variable iTable       as integer   no-undo.
  define variable hBuffer      as handle    no-undo.
  define variable cQuery       as character no-undo.
  define variable hQuery       as handle    no-undo. 
  define variable iDSbuffer    as integer   no-undo. 
  define variable oQueryString as class DSQueryString.
   
  run attachDataSources in target-procedure 
                           (input-output dataset-handle phDataset by-reference). 
 
  run defineReadEvents in target-procedure 
                           (input-output dataset-handle phDataset by-reference)
                           no-error.
                         
  do iBuffer = 1 to phDataset:num-buffers:
    assign /* set handle and check if table is requested */
      hBuffer = phDataSet:get-buffer-handle(iBuffer)
      iTable  = lookup(hBuffer:name,pcTables).
    /* if table found (and valid datasource )*/ 
    if iTable > 0 then
    do: 
      if valid-handle(hBuffer:data-source) then 
      do:
        if iTable = 1 then
        do: 
          /* set batchsize */
          hBuffer:batch-size = piBatchSize.
          /* if context set batch start */
          if pcContext > '' then 
            hBuffer:data-source:restart-rowid = to-rowid(pcContext).
          /* query mapping with callback to sourcecolumn and source-default-query */  
          cQuery  = entry(1,pcQueries,chr(1)).
          oQueryString = new DSQueryString
                        (cQuery, target-procedure, hBuffer:name).
          create query hQuery.
          do iDSbuffer = 1 to hBuffer:data-source:query:num-buffers:  
            hQuery:add-buffer(hBuffer:data-source:query:get-buffer-handle(iDsbuffer)).
          end.
          hBuffer:data-source:query = hQuery.    
          oQueryString:insertToQuery(hQuery).
          delete object oQueryString. 
        end. /* first table */               
      end.   
    end.
    else
      hBuffer:fill-mode = "NO-FILL".     
  end.
  
  phDataset:fill().
  /* set batch context for first table */
  hBuffer = phDataset:get-buffer-handle(entry(1,pcTables)).
  if not hBuffer:last-batch then
    pcContext = string(hBuffer:data-source:next-rowid).
  else
    pcContext = ''.  
  run detachDataSources in target-procedure
             (input-output dataset-handle phDataset by-reference).
  delete object hQuery.           
end. /* retrieveData */

PROCEDURE saveBuffer :
/*------------------------------------------------------------------------------
  Purpose:     Called from saveChanges to process changes
               for a single DataSet temp-table buffer.
  Parameters:  buffer handle
  Notes:       This is similar to the OERI BE version, but has no pre and
               post transaction hooks because:  
               - This method might very well be called inside a transaction.
               - The BE supports hooks outside of the transaction scope. 
------------------------------------------------------------------------------*/

    DEFINE INPUT PARAMETER phBuffer AS HANDLE NO-UNDO.

    DEFINE VARIABLE hBeforeBuff AS HANDLE NO-UNDO.
    DEFINE VARIABLE cLogicProc  AS CHARACTER NO-UNDO.
    DEFINE VARIABLE hDataSet    AS HANDLE     NO-UNDO.
    define variable hBeforeQry  as handle    no-undo.

    /* phBuffer is the after-table. Walk through the modified rows in
       the before-table, specifically to capture DELETEs, even though it
       is of course the after-table values that Progress writes to the db
       for CREATEs and MODIFYs. */
    hBeforeBuff = phBuffer:BEFORE-BUFFER.
    
    IF VALID-HANDLE(hBeforeBuff) THEN
    DO: 
      create query hBeforeQry.
      hBeforeQry:SET-BUFFERS(hBeforeBuff).
      hBeforeQry:QUERY-PREPARE("FOR EACH " + hBeforeBuff:NAME).
      hBeforeQry:QUERY-OPEN().
      hBeforeQry:GET-FIRST().
      /* Process all the updates for the current table. */
      DO WHILE NOT hBeforeQry:QUERY-OFF-END:
        RUN commitChanges IN target-procedure (INPUT phBuffer).
        hBeforeQry:GET-NEXT().
      END. /* END DO WHILE NOT QUERY-OFF-END */
    END. /* END DO IF VALID-HANDLE */
END PROCEDURE. /* saveBuffer */

procedure saveChanges:
/*------------------------------------------------------------------------------
  Purpose:  Save changes to physical storage                
Parameters: dataset dsOrder 
     Notes: This example of data access save traverses all records in order
            to: 
            - Allow the DA to control transactions, which cannot be directly 
              controlled in distributed BL. Transaction control is a concern of 
              the BL, but should not be contolled with TRANSACTION keyword.
            - Offer a course grained save interface from distributed BL    
          - The default operation here is just a transaction per record
          - This simple sample does not use child relations to traverse 
            the buffers as the OERI BE example. (A property that controls 
            update order would give the same effect).
          - Traversing of child data from a parent would make sense if done for 
            one parent record at a time.     
-----------------------------------------------------------------------------*/
  define input-output parameter dataset-handle phDataset.

  define variable lOk         as logical no-undo. 
  define variable iBuffer     as integer   no-undo.
  define variable hBuffer     as handle    no-undo.

  run attachDataSources in target-procedure
                           (input-output dataset-handle phDataset by-reference). 
  
  do transaction:
    do iBuffer = 1 TO phDataSet:num-buffers:
      hBuffer = phDataSet:get-buffer-handle(iBuffer).
      run saveBuffer in target-procedure(hBuffer).
    end.  /* do iBuffer = 1 */
  end.
  run detachDataSources in target-procedure
                           (input-output dataset-handle phDataset by-reference).  
end. /* saveChanges */ 

PROCEDURE commitChanges :
/*------------------------------------------------------------------------------
  Purpose: This procedure is called from saveChanges in the DA
           to do the actual database commit and any transaction-related
           validation logic for a buffer.      
  Parameters:  buffer handle.
  Notes:   This is taken directly from the OERI. 
        -  Note that in this sample that illustrates support for distributed 
           business logic the DA saves the whole dataset in SaveChanges, so 
           this is not currently called from the BE.  
        -  The event names are the same as in the OERI. The word "Trans" 
           is somewhat misleading as the DA need to support large transactions.
           One could consider using a name that indicates that this is a row 
           event instead.
------------------------------------------------------------------------------*/
    DEFINE INPUT  PARAMETER phBuffer     AS HANDLE     NO-UNDO.

    DEFINE VARIABLE hDataSet    AS HANDLE     NO-UNDO.
    DEFINE VARIABLE hBeforeBuff AS HANDLE     NO-UNDO.
    DEFINE VARIABLE cLogicProc  AS CHARACTER  NO-UNDO.

    ASSIGN hBeforeBuff = phBuffer:BEFORE-BUFFER
           hDataSet    = phBuffer:DATASET
           cLogicProc  = phBuffer:TABLE-HANDLE:NAME 
                     + IF hBeforeBuff:ROW-STATE = ROW-DELETED THEN "Delete"
                       ELSE IF hBeforeBuff:ROW-STATE = ROW-CREATED THEN "Create"
                       ELSE "Modify".
    /* This is a sub transaction, the actual transaction is in savechanges */                  
    DO TRANSACTION ON ERROR UNDO, LEAVE: 
      RUN VALUE (cLogicProc + "BeginTrans") IN TARGET-PROCEDURE
            (INPUT DATASET-HANDLE hDataSet BY-REFERENCE) NO-ERROR.
         
      IF NOT hBeforeBuff:ERROR THEN
      DO:
        hBeforeBuff:SAVE-ROW-CHANGES() no-error.
        IF NOT hBeforeBuff:ERROR THEN
          RUN VALUE (cLogicProc + "EndTrans") IN TARGET-PROCEDURE
              (INPUT DATASET-HANDLE hDataSet BY-REFERENCE) NO-ERROR.
        else do:
          hBeforeBuff:error-string = if return-value > '' 
                                     then return-value
                                     else error-status:get-message(1).
        end.
      END. /* END DO IF no error after BeginTrans validation. */       
    END.     /* END RowTrans TRANSACTION block */
END PROCEDURE.   /* commitChanges */  

procedure detachDataSources:
  define input-output parameter dataset-handle phDataset.
  define variable hBuffer as handle    no-undo.
  define variable iBuffer as integer   no-undo.
  do iBuffer = 1 to phDataset:num-buffers:
    hBuffer = phDataset:get-buffer-handle(iBuffer).
    if valid-handle(hBuffer:data-source) then 
      hBuffer:detach-data-source(). 
  end.
end. 
  
procedure createObject:
  /* nothing yet. Could create property storage.. see beSupport  */
end. 

procedure destroyObject:
  delete procedure target-procedure.  
end.     

/* used by locateProcedure (see locateprocedure.i) */
procedure isRunning:
  define output parameter phThis as handle no-undo.
  phThis = this-procedure.
end.  
       
       