/*************************************************************/
/* Copyright (c) 2007 by Progress Software Corporation      */
/*************************************************************/
/*------------------------------------------------------------------------------
    File   :   adm2/serviceadapter.p
    Purpose:   Data request service adapter interface
    Notes    - This is a sample/prototype that illustrates adaptations to a 
               service that only keeps one entity persistent
                - fetchData keeps track of the requests in a ttDataset TT 
                - stopEntity calls stop in service when last ttDataset resigns 
             - The adapter is responsible of starting and closing the service                  
             - Service requirements:   
               - BEs must support dynamic data transport 
                (static data prevents BE from supporting multiple instances)
               - Should have API to stop BE   
------------------------------------------------------------------------------*/
/* ************************  Definitions  ********************** */

define temp-table ttDataSet no-undo
  field Entity as character
  field DataSetHandle as handle
  index datasethandle as unique datasethandle
  index Entity Entity.    

define variable ghService as handle no-undo.

/* ************************  Function Prototypes ********************** */
function newDataset returns handle private  
  ( pcEntity as character) forward.
   
function getObjectType returns character 
  (  )  forward .

function getService returns handle private 
  (  )  forward .

function destroyService returns logical private 
  (  )  forward .


/* ***********************  Procedures ********************************* */
procedure destroyObject :
/*------------------------------------------------------------------------------
  Purpose:  Destroy     
------------------------------------------------------------------------------*/
  destroyService(). 
  delete procedure this-procedure .
end procedure .

procedure retrieveData :
/*------------------------------------------------------------------------------
  Purpose:  Retrieve data or definitions for a single Business Entity
 Features:  Single entity Adapter, relies on requester issuing separate 
            requests. (ADM2 - DataView override isFetchedByParent)     
Parameters:  
     pcEntity       - Business Entity 
     pcTables       - Tables for which to retrieve data or definitions
     pcQueries      - List of queries that corresponds to the requested tables
     pcJoins        - List of join fields for the tables
     pcPositions    - How to find a single record in the tables  
     pcRequest      - The actual request for the tables   
     pcBatchContext - Context info for a batch request 
     plFillBatch    - Read a full batch size at end of result set
     pcNumRecords   - Batch size for the tables               
     phDataSet      - Prodataset handle 
     pcContext      - Other context    
     pcPrevContext  - Context for prev record that was NOT returned from tables
     pcNextContext  - Context for next record that was NOT returned from tables
                                        
  Notes: The comments has been minimized to focus on code.
         (see orignal or whitepaper) 
     -   This is an example of a Service Adapter that ignores the request 
         for more than one Entity.
         This relies on an DataView override of isFetchedByParent that 
         just returns getDataIsFetched. This will ensure that the child entity
         is requested separately  (The default version figures out 
         if the table and entity was retrieved by the parent based on logging
         of the request tree)      
------------------------------------------------------------------------------*/
 define input         parameter pcEntity       as character no-undo extent.
 define input         parameter pcTables       as character no-undo extent.
 define input         parameter pcQueries      as character no-undo extent.
 define input         parameter pcJoins        as character no-undo extent.
 define input         parameter pcPositions    as character no-undo extent.
 define input         parameter pcRequests     as character no-undo extent.
 define input         parameter pcBatchContext as character no-undo.
 define input         parameter plFillBatch    as logical   no-undo.
 define input-output  parameter pcNumRecords   as character no-undo extent.
 define input-output  parameter phDataset      as handle    no-undo extent. 
 define input-output  parameter pcContext      as character no-undo extent.
 define output        parameter pcPrevContext  as character no-undo extent.
 define output        parameter pcNextContext  as character no-undo extent.
 
  /* dim output from input */
  extent(pcPrevContext) = extent(pcEntity).
  extent(pcNextContext) = extent(pcEntity).
 
  define variable hEntity     as handle    no-undo.
  define variable iEntity     as integer   no-undo.
  define variable hDataset    as handle    no-undo.
  define variable lNew        as logical   no-undo.
  define variable cContext    as character no-undo. 
  define variable iNumRecords as integer   no-undo.
  define variable iBuffer     as integer   no-undo.
  define variable iTable      as integer   no-undo.
  
  do iEntity = 1 to extent(pcEntity):   
    /* input batchcontext only applies to first entity */
    if iEntity = 1 then
      cContext = pcBatchContext.
    else 
      cContext = ''. 
     
     /* This service only batches first table */
    iNumRecords = integer(entry(1,pcNumRecords[iEntity])).
 
    /* find not supported in service ... read all  */
    if pcRequests[iEntity] begins 'WHERE ' then 
      iNumRecords = 0.
    
    /* not valid dataset means request for new instance (including defs) */  
    if not valid-handle(phDataset[iEntity]) then
    do:
      /* We call specific method in the service that returns the 
	     dataset if it is in the session and make a dynamic instance of it
	     that we pass to fetch 
	    (with output only it will be necessary to run one BE instance 
	     for each requested instance )
	    Returns ? if the BE is dynamic */ 
      lNew = true.  
      phDataset[iEntity] = newDataset(pcEntity[iEntity]).
    end.
       
    if pcTables[iEntity] <> '' or not valid-handle(phDataset[iEntity]) then
    do:  
      run fetchData in getService() (pcEntity[iEntity],
                                     pcTables[iEntity],
                                     pcQueries[iEntity],
                                     pcRequests[iEntity],
                                     iNumRecords,
                                     input-output cContext,
                                     input-output phDataset[iEntity])
                                     no-error.
   
      if error-status:error then return error if return-value > '' then return-value else error-status:get-message(1).                                   
      /* This service only returns context for first table 
        ADM2 will handle fewer entries than tables */
      if iEntity = 1 then
        pcNextContext = cContext.
    end.
      /* keep track of requested instances */
    if lNew then 
    do:
      /* Not used as dataset, but as index to identify adm2 stopEntity */
      create ttDataset.
      assign
        ttDataset.Entity = pcEntity[iEntity]
        ttDataset.DataSetHandle = phDataset[iEntity]
        .     
    end.     
  end. /* do iEntity = 1 */          
end procedure /* retrieveData */. 

procedure submitData :
/*------------------------------------------------------------------------------
  Purpose:     Submit of changes
  parameters:  
     pcEntity       - Business Entity 
     phDataSet      - Prodataset handle with changes only 
     pcContext      - Other context    
     Notes:        
------------------------------------------------------------------------------*/
  /* Business Entity reference */
  define input         parameter pcEntity     as character no-undo extent.
  define input-output  parameter phDataset    as handle    no-undo extent.
  define input-output  parameter pcContext    as character no-undo extent.

  /* we don't loop as we know the adm2 does not submit more than one */  
  run saveChanges in getService()(pcEntity[1],
                                  input-output phDataset[1]) no-error. 
  if error-status:error then  
    return error if return-value > '' then return-value else error-status:get-message(1).  
  
  run markRows ( phDataset[1],'').
  return.                              
end procedure .

/* receive message that adm2 is done (the data set is already deleted) */
procedure stopEntity:
  define input parameter phDataSet as handle.
  define variable cEntity as character no-undo.
  find ttDataset where ttDataset.DataSetHandle = phDataset no-error.
  if available ttDataset then 
  do:
    cEntity = ttDataset.Entity.
    delete ttDataset.
    /* signal service if this is last dataset instance */ 
    if not can-find(first ttDataSet where ttDataSet.Entity = cEntity) then
      run stopEntity in getService() (cEntity).
  end.  
end.

procedure markRows:
  /* marks all rows as REJECTED or ERROR 
     Use when the service does not do this with large transactions */ 
  define input parameter phDataSet as handle    no-undo.
  define input parameter pcFlag    as character no-undo.
  
  define variable iBuf   as integer no-undo.
  define variable hBuf   as handle  no-undo.
  define variable hBefore as handle  no-undo.
  
  define variable hQuery as handle no-undo.
  
  create query hQuery.
  do ibuf = 1 to phDataSet:num-buffers:
    hBuf = phdataset:get-buffer-handle(iBuf).
    hBefore = hBuf:before-buffer.
    hQuery:set-buffers(hBefore).
    hQuery:query-prepare("FOR EACH " + hBefore:name).
    hQuery:query-open().
    hQuery:get-first().
    do while not hQuery:query-off-end :
      if hBefore:error = false and hBefore:rejected = false then
      do:
        if pcFlag begins 'reject' then
          hBefore:rejected = true.
        else
          hBefore:error = true.
      end.
      hQuery:get-next().
    end. /* do while not off end */ 
  end. /* ibuf loop */
  delete object hQuery.
end. /* procedure markRows */
  
/* ************************  Function Implementations ***************** */

/* Returns a new empty dataset instance  
  (we want dynamic instances in order to use a single BE for sll)  */
function newDataset returns handle private  
  ( pcEntity as character):
  define variable hDataset as handle no-undo. 
  define variable cContext as character no-undo.
  define variable hNewDataset as handle no-undo. 
  
  run fetchDataset in getService() (pcEntity,output hDataset).
  if valid-handle(hDataset) then
  do:                               
    create dataset hNewDataset.   
    hNewDataset:create-like(hDataset).   
  end.   
  return hNewDataset.
end.

FUNCTION getObjectType RETURNS CHARACTER
  (  ) :
/*------------------------------------------------------------------------------
  Purpose: To be used by designtime super destroy  
    Notes:  
------------------------------------------------------------------------------*/
  RETURN "Service". 
END FUNCTION.

/* get service starts it on first call  */
function getService returns handle private 
  (  ):
  if not valid-handle(ghService) then 
  do on stop undo, leave:
    run support/service.p persistent set ghService.  
    /* tell service to set HasClient in started BE */
    dynamic-function("setOperatingMode" in ghService,'client'). 
  end.
  return ghService. 
end.

function destroyService returns logical private 
  (  ):
  define variable h as handle    no-undo.  
  if valid-handle(ghService) then 
  do:
    for each ttDataset break by ttDataset.Entity:
      if first-of(ttDataset.Entity) then 
        run stopEntity in getService() (ttDataset.Entity).
    end.  
    run destroyObject in ghService.
    return true.
  end.   
  return false. 
end.
