/*------------------------------------------------------------------------------
 File        : daOrder order data access   
 Description : Example/prototype code with all static and no super code
 Notes       : Illustrates solution for two issues 
               1. Logical data wants real key  
               2. Database defines view-as with hard coded list for creditcard
               we return it normalized as table with extrenal keys (pretend).
             - fetchData supports conditional retrieval of tables
             - All request methods are statefree (attaches and detaches sources) 
               and passes dataset around  by-reference. (no get or set)
             - attach and detach methods with no event definitions 
             - Implements saveChanges with each order and child orderlines
               in same transaction. 
             - saveChanges sets rejected flag in orderlines backed out of 
               transaction.          
 ------------------------------------------------------------------------------*/

/* ***************************  Definitions  ************************** */
/* defines dataset include and loads daSupport as super */ 
{support/daEntity.i Order} 

define data-source dOrder for Order.
define data-source dOrderLine for OrderLine.
define data-source dItem for Item.

define variable gcCreditCardKeys as character no-undo
  init 'AmEx,MC,Visa,Disc'. 
define variable gcCreditCardNames as character no-undo
  init 'American Express,Master Card,Visa,Discover Card'. 
 
/* ************ Functions ****************************************/           
/* call backs for query mapping */
function sourceDefaultQuery returns character  
  (pcTable as char):
  case pcTable:
    when 'eOrder' then 
      return 'For each Order indexed-reposition'.
    when 'eOrderLine' then 
      return 'For each OrderLine indexed-reposition'.
  end.    
end.

function sourceColumn returns character  
  (pcTable as char,
   pcColumn as char ):
  define variable cMap  as character no-undo.
  case pcTable:
    when 'eOrder' then
    do: 
      cMap = data-source dorder:get-dataset-buffer:data-source-complete-map.
    end.  
    when 'eOrderLine' then
    do: 
      cMap = data-source dorderline:get-dataset-buffer:data-source-complete-map.
    end.  
  end.
  return entry(lookup(pccolumn,cMap) + 1,cMap).
end.

function creditCardName returns character private 
  (pcKey as character):
  return entry(lookup(pcKey,gcCreditCardKeys),gcCreditCardNames).
end.

function creditCardKey returns character private 
  (pcName as character):
  return entry(lookup(pcName,gcCreditCardNames),gcCreditCardKeys).
end.

function errorString returns character private 
  (pcString as char):
  if pcString > '' then 
    return pcString.
  else    
    return error-status:get-message(1). 
end.
   
/* ************ Procedures ****************************************/        
/* Called from fetchData */ 
procedure defineReadEvents : 
  define input-output parameter dataset for dsOrder.
 
  buffer eOrder:set-callback-procedure
            ("AFTER-ROW-FILL", "eOrderAfterRowFill", target-procedure).
  
  buffer eCreditCard:set-callback-procedure
            ("BEFORE-FILL", "fillCreditCard", target-procedure).     

end.

procedure attachDataSources:
  define input-output parameter dataset for dsOrder.
  
  buffer eOrder:attach-data-source(data-source dOrder:handle,'','CreditCard').        
  buffer eOrderLine:attach-data-source(data-source dOrderLine:handle).
  buffer eItem:attach-data-source(data-source dItem:handle).  
end.   

procedure saveChanges:
/*------------------------------------------------------------------------------
  Purpose:  Save changes to physical storage (completely overrides daSupport)               
Parameters: dataset dsOrder 
     Notes: This example of data access save traverses all records in order
            to control transactions, which cannot be directly controlled in BL.
            (at least not in distributed BL )
-----------------------------------------------------------------------------*/
  define input-output parameter dataset for dsOrder.

  define variable lOk         as logical no-undo.  
  
  run attachDataSources (input-output dataset dsOrder by-reference). 
  
  /* traverse before records to get deletes */  
  for each eOrderBefore:
    lOk = false.
    OrderTrans:
    do transaction:
      if buffer eOrderBefore:row-state <> row-created then
      do:
        find order where order.ordernum = eOrderBefore.ordernum exclusive no-error.
        if not available order then
        do: 
          assign  /* (UNTESTED)*/
            buffer eOrderBefore:error = true
            buffer eOrderBefore:data-source-modified = true.
            UNDO OrderTrans,LEAVE OrderTrans.
        end. /* not avail or locked order */
        /* optimistic lock check for transformed data (UNTESTED)*/
        if Order.Creditcard <> creditCardName(eOrderBefore.CreditCard) then 
        do:
          assign 
            buffer eOrderBefore:error = true
            buffer eOrderBefore:data-source-modified = true.
          run eOrderAfterRowFill. /* refresh */
          /* should ideally not undo here, but continue save to get other 
             possible lock conflict data to client  */
          UNDO OrderTrans,LEAVE OrderTrans. 
        end.
        /* sanity check for transformed data (UNTESTED) */
        find eOrder where rowid(eOrder) = buffer eorderbefore:after-rowid.         
        if creditCardName(eOrder.CreditCard) = ? then 
        do: 
          assign 
            buffer eOrderBefore:error = true
            buffer eOrderBefore:error-string = 'Invalid creditcard'.    
            UNDO OrderTrans,LEAVE OrderTrans.
        end.              
        /* do orderline deletes first */
        for each eOrderLineBefore 
            where eOrderLineBefore.Ordernum = eOrderBefore.ordernum
            and buffer eOrderLineBefore:row-state = row-deleted:             
          buffer eOrderLineBefore:save-row-changes() no-error.
          if buffer eOrderBefore:error then 
          do:
            buffer eOrderBefore:error-string = errorString(return-value).
            UNDO OrderTrans,LEAVE OrderTrans.
          end.  
        end.          
      end. /* not new */ 
      find eOrder where rowid(eOrder) = buffer eorderbefore:after-rowid.         
      find Salesrep where Salesrep.SalesRep = eOrder.salesrep no-error.
      if not available Salesrep then
      do: 
        buffer eOrder:error-string = 'Invalid Salesrep'.
        buffer eOrder:error = true.
        UNDO OrderTrans,LEAVE OrderTrans.
      end.
      /* Save */
      buffer eOrderBefore:save-row-changes(1,'ordernum') no-error.      
      if buffer eOrderBefore:error then
      do: 
        buffer eOrderBefore:error-string = errorString(return-value).
        /* refresh transformated data if lock conflict */
        if buffer eOrderBefore:data-source-modified then
          run eOrderAfterRowFill.
        UNDO OrderTrans,LEAVE OrderTrans.
      end.
      
      /* transform Creditcard to internal format */
      find eOrder where rowid(eOrder) = buffer eorderbefore:after-rowid.         
      find order where order.ordernum = eOrder.ordernum exclusive.
      Order.CreditCard = creditCardName(eOrder.CreditCard).    
      
      /* save orderlines */
      for each eOrderLineBefore 
          where eOrderLineBefore.Ordernum = eOrderBefore.ordernum         
            and buffer eOrderLineBefore:row-state <> row-deleted:             
        buffer eOrderLineBefore:save-row-changes() no-error.
        if buffer eOrderBefore:error then 
        do:
          buffer eOrderBefore:error-string = errorString(return-value).
          UNDO OrderTrans,LEAVE OrderTrans.
        end.
      end.    
      /* set flag for completed trans */
      lOk = true.
    end. /* transaction */
    /* if failed transaction set rejected flag on rolled back records */
    if not lOk then 
    do:
      dataset dsOrder:error = true.
      if not buffer eOrderBefore:error then 
        buffer eOrderBefore:rejected = true.
      for each eOrderLineBefore 
          where eOrderLineBefore.Ordernum = eOrderBefore.ordernum:             
        if not buffer eOrderLineBefore:error then 
          buffer eOrderLineBefore:rejected = true.
      end. /* for each eOrderLineBefore */          
    end.  /* not ok */
  end. /* for each eOrderBefore */  
 
  /* take care of standalone orderline changes 
    (one trans per line, should possibly have one per ordernum here also) */ 
  for each eorderLineBefore 
      where not can-find(eOrderBefore where eOrderBefore.Ordernum = eOrderLineBefore.ordernum)
  transaction:      
    buffer eOrderLineBefore:save-row-changes() no-error.
    if buffer eOrderLineBefore:error then 
    do:
      dataset dsOrder:error = true.
      buffer eOrderLineBefore:error-string = errorString(return-value).
      UNDO,NEXT.
    end.      
  end. /* for each eOrderLineBefore transaction */
  run detachDataSources (input-output dataset dsOrder by-reference). 
end. /* saveChanges */ 
 
procedure fillCreditCard:
  define input parameter dataset for dsOrder.
  /* Physical data do not normalize creditcard info, use internal key 
     and depend on view-as. This illustrates a solution for two issues 
     1. Logical data wants real key (pretend: data is from external register) 
     for interoperability with other (services...) 
     2. A UI cannot be expected to support view-as and should not need to 
        hardcode list, but should be able to bind to a table          */    
  define variable iCard  as integer   no-undo.
  do iCard = 1 to num-entries(gcCreditCardKeys):
    create eCreditCard.
    assign
      eCreditCard.CreditCard     = entry(iCard,gcCreditCardKeys)
      eCreditCard.CreditCardName = entry(iCard,gcCreditCardNames).
  end.           
end.  /* fillCreditCard */    

procedure eOrderAfterRowFill:  
  define input parameter dataset for dsOrder.
  /* transform internal physical key to external key  
     This is not meant to indicate that this is desirable, but highlight that 
     the DA layer can overcome physical limitations  
     - keep in mind that any row level hook add cost to the fill */       
  eOrder.CreditCard = creditCardKey(Order.CreditCard).
end.  /* eOrderAfterRowFill */   


       