/*------------------------------------------------------------------------------
 File        : daOrder order data access   
 Description : Example/prototype code with all static and no super code
 Notes       : Illustrates solution for two issues 
               1. Logical data wants real key  
               2. Database defines view-as with hard coded list for creditcard
               we return it normalized as table with extrenal keys (pretend).
             - fetchData supports conditional retrieval of tables
             - All request methods are statefree (attaches and detaches sources) 
               and passes dataset around  by-reference. (no get or set)
             - attach and detach methods with no event definitions 
             - Implements saveChanges with each order and child orderlines
               in same transaction. 
             - saveChanges sets rejected flag in orderlines backed out of 
               transaction.          
 ------------------------------------------------------------------------------*/

/* ***************************  Definitions  ************************** */
/* defines dataset include and loads daSupport as super */ 

{schema/dsorder.i} 

define data-source dOrder for Order.
define data-source dOrderLine for OrderLine.
define data-source dItem for Item.

define variable gcCreditCardKeys as character no-undo
  init 'AmEx,MC,Visa,Disc'. 
define variable gcCreditCardNames as character no-undo
  init 'American Express,Master Card,Visa,Discover Card'. 
 
/* ************ Functions ****************************************/           
 
function creditCardName returns character private 
  (pcKey as character):
  return entry(lookup(pcKey,gcCreditCardKeys),gcCreditCardNames).
end.

function creditCardKey returns character private 
  (pcName as character):
  return entry(lookup(pcName,gcCreditCardNames),gcCreditCardKeys).
end.

function errorString returns character private 
  (pcString as char):
  if pcString > '' then 
    return pcString.
  else    
    return error-status:get-message(1). 
end.
   
/* ************ Procedures ****************************************/        
/* Called from fetchData */ 
procedure defineReadEvents : 
  define input-output parameter dataset for dsOrder.
 
  buffer eOrder:set-callback-procedure
            ("AFTER-ROW-FILL", "eOrderAfterRowFill", target-procedure).
  
  buffer eCreditCard:set-callback-procedure
            ("BEFORE-FILL", "fillCreditCard", target-procedure).     

end.

procedure attachDataSources:
  define input-output parameter dataset for dsOrder.
  
  buffer eOrder:attach-data-source(data-source dOrder:handle,'','CreditCard').        
  buffer eOrderLine:attach-data-source(data-source dOrderLine:handle).
  buffer eItem:attach-data-source(data-source dItem:handle).  
end.   

procedure fetchData:
/*------------------------------------------------------------------------------
  Purpose: Fetch data from the physical storage.
               
Parameters:  
     phDataSet      - Prodataset handle (NOT dataset-handle)
     pcContext      - context    
     pcTables       - Tables for which to retrieve data or definitions
     pcQueries      - List of queries that corresponds to the requested tables
     pcRequest      - The actual request for the tables   
     piNumRecords   - Batch sizes
 
   Notes:  uses static data sources   
 -----------------------------------------------------------------------------*/
  define input        parameter pcTables       as character no-undo.
  define input        parameter pcQueries      as character no-undo.
  define input-output parameter dataset-handle phDataset.
  
  define variable iBuffer     as integer   no-undo.
  define variable iTable      as integer   no-undo.
  define variable hBuffer     as handle    no-undo.
  define variable hQuery      as handle    no-undo. 
  define variable hParentRel  as handle    no-undo. 
       
  run attachDataSources in target-procedure 
                           (input-output dataset-handle phDataset by-reference). 
  
  run defineReadEvents in target-procedure 
                           (input-output dataset-handle phDataset by-reference)
                           no-error.
                          
  do iBuffer = 1 to phDataset:num-buffers:
    assign /* set handle and check if table is requested */
      hBuffer  = phDataSet:get-buffer-handle(iBuffer)
      iTable   = lookup(hBuffer:name,pcTables).
    /* if table found (and valid datasource )*/ 
    if iTable > 0 then
    do: 
  
        if valid-handle(hBuffer:data-source) then 
        do:
          if iTable = 1 then
          do: 
            hParentRel = hBuffer:PARENT-RELATION.
            /* if child then turn off parent relation as this is a request for 
               the child only */
            if valid-handle(hParentRel) then 
              hParentRel:active = false.
              
           
          end.
          /* query mapping (callback to sourcecolumn and source-default-query)*/  
          hQuery = hBuffer:data-source:query.
          hQuery:query-prepare(pcqueries).
          hQuery:query-open().
             
        end.
    end.
    else
      hBuffer:fill-mode = "NO-FILL".     
  end.
  /*if phDataset::eorderline:data-source:query:prepare-string > '' then.*/
  phDataset:fill().
end.
 
procedure fillCreditCard:
  define input parameter dataset for dsOrder.
  /* Physical data do not normalize creditcard info, use internal key 
     and depend on view-as. This illustrates a solution for two issues 
     1. Logical data wants real key (pretend: data is from external register) 
     for interoperability with other (services...) 
     2. A UI cannot be expected to support view-as and should not need to 
        hardcode list, but should be able to bind to a table          */    
  define variable iCard  as integer   no-undo.
  do iCard = 1 to num-entries(gcCreditCardKeys):
    create eCreditCard.
    assign
      eCreditCard.CreditCard     = entry(iCard,gcCreditCardKeys)
      eCreditCard.CreditCardName = entry(iCard,gcCreditCardNames).
  end.           
end.  /* fillCreditCard */    

procedure eOrderAfterRowFill:  
  define input parameter dataset for dsOrder.
  /* transform internal physical key to external key  
     This is not meant to indicate that this is desirable, but highlight that 
     the DA layer can overcome physical limitations  
     - keep in mind that any row level hook add cost to the fill */       
  eOrder.CreditCard = creditCardKey(Order.CreditCard).
end.  /* eOrderAfterRowFill */   


       