/*************************************************************/
/* Copyright (c) 2007 by Progress Software Corporation      */
/*************************************************************/
/*------------------------------------------------------------------------------
 File        : beSupport.p
 Description : Business Entity super procedre 
 Notes       : 
              -  Service as property (default to source in main) 
              -  DataAccess start and stop managed by service  
              -  This illustrates use of hooks to separate read and 
                 save business logic from the actual data transport and 
                 context. This allows future changes as well as replacements 
                 of the fetchData and saveChanges methods.   
              -  This sample only uses one hook at the dataset level. 
                 One should consider adding table and row level hooks, but 
                 these hooks need to be spearated from similar hooks in the DA
                 to keep the BusinessEntity distributable. 
              -  This sample illustrates hooks in the instance. Hooks can also 
                 be in separate procedures.                                               
------------------------------------------------------------------------------*/

/* ***************************  Definitions  ************************** */
/* property table -- 
   - Stores instance properties. 
    (consider using classes with native properties instead)*/                                 
define temp-table ttProperty no-undo
  field TargetProcedure as handle
  field EntityName as character 
  field Service   as handle 
  field DataAccessObject as handle
  field HasClient as logical
  index targetprocedure as unique targetprocedure.  

/* ***************************  ProtoTypes **************************** */
function findTarget returns logical private
  ( phTarget as handle ) forward.

/* ************ Functions ****************************************/                           
/* find the target's properties  - private   */ 
function findTarget returns logical private
  ( phTarget as handle ):
  
  find ttProperty where ttProperty.targetprocedure = phTarget no-error.
  
  return available ttProperty.                                           
end.       

function getDataAccessObject returns handle 
  (  ):
  define variable hService as handle    no-undo.

  findTarget(target-procedure).
  if not valid-handle(ttProperty.DataAccessObject) then
  do:
    hService = dynamic-function('getService' in target-procedure). 
    ttProperty.DataAccessObject = dynamic-function('startDataAccess' in hService,
                                  target-procedure).
  end.
  return ttProperty.DataAccessObject.                                           
end.  
     
function setDataAccessObject returns logical private
  ( phHandle as handle ) :
  findTarget(target-procedure).
  ttProperty.DataAccessObject = phHandle.                                           
end.  

/* HasClient is set to true if the Be is directly supporting a rich client 
  (This basically means define row level triggers and keep alive BE ) */ 
     
function getHasClient returns logical  
  (  ) .
  findTarget(target-procedure).
  return ttProperty.HasClient.                                           
end.

function setHasClient returns logical 
  ( plClient as logical ) :
  findTarget(target-procedure).
  ttProperty.HasClient = plClient.
  return true.                                           
end.

/* EntityName is defined here for dynamic instances  */      
function getEntityName returns character  
  (  ): 
  findTarget(target-procedure).
  return ttProperty.EntityName.                                           
end.

function setEntityName returns logical 
  ( pcName as character ) :
  findTarget(target-procedure).
  ttProperty.EntityName = pcName.
  return true.                                           
end.

function getService returns handle 
  (  ) .
  findTarget(target-procedure).
  return ttProperty.Service.                                           
end.

function setService returns logical 
  ( phService as handle ) :
  findTarget(target-procedure).
  ttProperty.Service = phService.
  return true.                                           
end.

/* Overridden by static instance */
function getStaticDataset returns handle ( ) :
  return ?.                                           
end.

/* To be used by generic super destroy  */
function getObjectType returns character ( ) :
  if target-procedure = this-procedure then 
    return "super". 
  else 
    return "BusinessEntity".                                            
end.
        
/* ************ Procedures ****************************************/        
procedure createObject:
  create ttProperty. 
  ttProperty.TargetProcedure = target-procedure.  
end.     

procedure destroyObject:
  define variable hService as handle    no-undo.
  hService = dynamic-function('getService' in target-procedure). 
  dynamic-function('stopDataAccess' in hService,target-procedure). 
  findTarget(target-procedure).
  delete ttProperty.
  delete procedure target-procedure.  
end.     

procedure fetchDataset:
/*------------------------------------------------------------------------------
  Purpose:  Fetch dataset  
 -----------------------------------------------------------------------------*/
  define output parameter dataset-handle hdataset.
  hDataset = dynamic-function('getStaticDataset' in target-procedure).  
end. /* fetchData */ 

procedure fetchData:
/*------------------------------------------------------------------------------
  Purpose:  Fetch data and/or definitions  
               
Parameters:  
     pcTables       - Tables for which to retrieve data or definitions
     pcQueries      - List of queries that corresponds to the requested tables
     pcRequest      - The actual request for the tables   
     pcNumRecords   - Batch sizes
     phDataSet      - Prodataset handle (not dataset-handle) 
     pcContext      - context
         
 -----------------------------------------------------------------------------*/
  define input        parameter pcTables       as character no-undo.
  define input        parameter pcQueries      as character no-undo.
  define input        parameter pcRequests     as character no-undo.
  define input        parameter piNumRecords   as integer   no-undo.
  define input-output parameter pcContext      as character no-undo.
  define input-output parameter dataset-handle phDataset.
  
  define variable hAccess  as handle    no-undo.
  define variable cName    as character no-undo. 
  define variable iBuffer  as integer   no-undo. 
   
  /* don't fetch for design time request */
  if pcTables > '' and pcRequests <> 'DEFS' then
  do: 
    hAccess = dynamic-function('getDataAccessObject' in target-procedure).
    run fetchData in hAccess
                    (pcTables,
                     pcQueries,
                     pcRequests,
                     piNumRecords,
                     input-output pcContext,
                     input-output dataset-handle phDataset by-reference) no-error.
    
    if error-status:error then 
      return error if return-value > '' then return-value else error-status:get-message(1).                  
    /* workaround: the adm2 does not turn off tracking changes 
       (it will be turned back on)*/               
    do iBuffer = 1 to phDataSet:num-buffers:
      phDataSet:get-buffer-handle(iBuffer):table-handle:tracking-changes = false.  
    end.  
    /* receive hook */
    run receiveData in target-procedure(input-output dataset-handle phDataset by-reference)
        no-error.
           
    if dynamic-function('getHasClient' in target-procedure) then
      run defineRowEvents in target-procedure 
                           (input-output dataset-handle phDataset by-reference)
                           no-error
                         .
  end. /* runtime request */                                          
end. /* fetchData */ 

procedure saveChanges:
/*------------------------------------------------------------------------------
   Purpose: Save data                
Parameters: phDataSet      - Prodataset with changes   
----------------------------------------------------------------------------*/
  define input-output parameter dataset-handle phDataset.
  
  define variable hAccess  as handle    no-undo.
 
  hAccess = dynamic-function('getDataAccessObject' in target-procedure).
  
  /* save hook */
  run validateData in target-procedure(input-output dataset-handle phDataset by-reference)
      no-error.
  if not phDataset:error then
  do: 
    run saveChanges in hAccess(input-output dataset-handle phDataset by-reference) 
        no-error. 
    if error-status:error then 
      return error if return-value > '' then return-value else error-status:get-message(1).
    /* receive hook */
    run receiveData in target-procedure(input-output dataset-handle phDataset by-reference)
        no-error.     
  end.                              
end.  

/* used by locateProcedure (see locateprocedure.i) */
procedure isRunning:
  define output parameter phThis as handle no-undo.
  phThis = this-procedure.
end.  
       
       
       